#include <X11/Xlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include "pixmap.h"
#include "globals.h"
#include "menu.h"
#include "menuhandler.h"
#include "font.h"

extern LPFONT pFont;

#define TRIANGLE_WIDTH 4
#define TRIANGLE_HEIGHT 8

#ifndef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#endif

#define MENU_OFFSET 4
#define MENU_SEPARATOR_HEIGHT 8

typedef struct _tagSMMI
{
  char *name;			/* if name is NULL, it will be a separator */
  char *args;
  int freeIcon;
  PIXMAP icon;
  /* options */
  int useIcons;
  MENUSTYLE ms;
} SMMI,*LPSMMI;

static void Launch(LPSMMI pParam)
{
  char *argv[4];
  
  argv[0] = "/bin/sh";
  argv[1] = "-c";
  argv[2] = pParam->args;
  argv[3] = 0;

  if (fork() == 0)
  {
    /* now close output and input */
    close(0);
    close(1);
    close(2);

    open("/dev/null",O_RDWR);
    dup(0); 
    dup(0); 
    
    execve(argv[0],argv,globalEnv);
    /* should not reach this */
    printf("Failed to execute program: %s\n",pParam->args);
    exit(-99);
  }
}

static void LaunchProgram(LPMENUOPERATION lpmo,LPSMMI pParam)
{
  FONTRECT fr;
  int len;
  int x,y;
  int renderX;
  int imageY;
  int imageBorder = 1;

  switch (lpmo->cmd)
  {
  case MENU_OP_GETSIZE:
    /* if a separator */
    if ((lpmo->data.size.drawFlags & MENU_DF_SEPARATOR) || !pParam->name)
    {
      lpmo->data.size.width = 2 * MENU_OFFSET;
      lpmo->data.size.height = MENU_SEPARATOR_HEIGHT;
    }
    else			/* a normal item */
    {
      pFont->GetExtents(pFont,pParam->name,strlen(pParam->name),&fr); 
      lpmo->data.size.width = fr.width + 2*MENU_OFFSET;
      lpmo->data.size.height = max(pFont->ascent,fr.height) + 2*MENU_OFFSET;
      /* spacing for icons */
      if (pParam->icon.icon != None)
      {
	lpmo->data.size.width += pParam->icon.width + 2*(1+imageBorder) + MENU_OFFSET;
	if (lpmo->data.size.height < (pParam->icon.height + 2*(1+imageBorder)))
	{
	  lpmo->data.size.height = pParam->icon.height + 2*(1+imageBorder);
	}
      }
      if (lpmo->data.size.drawFlags & MENU_DF_SUBMENU)
      {
	lpmo->data.size.width += TRIANGLE_WIDTH + 4;
	if (lpmo->data.size.height < TRIANGLE_HEIGHT)
	{
	  lpmo->data.size.height = TRIANGLE_HEIGHT + 2;
	}
      }
    }
    break;
  case MENU_OP_CLICKED:
    /* printf("I've been clicked %s\n",pParam->name); */
    if (pParam->args && *pParam->args)
    {
      Launch(pParam);
    }
    break;
    /*case MENU_OP_SELECTED:
    printf("I've been selected %s\n",pParam->name);
    break;
  case MENU_OP_UNSELECTED:
    //printf("I've been unselected %s\n",pParam->name);
    break; */
  case MENU_OP_RENDER:
    /* if a separator */
    if ((lpmo->data.draw.drawFlags & MENU_DF_SEPARATOR) || !pParam->name)
    {
      /* precalculate some useful variables */
      y = lpmo->data.draw.y + lpmo->data.draw.height/2 - 1;
      x = lpmo->data.draw.x + 1;
      renderX = lpmo->data.draw.x + lpmo->data.draw.width - 2;
      
      /* now render */
      XSetForeground(pDisp,gGc,PALETTE_DARK_GRAY);
      XDrawLine(pDisp,lpmo->data.draw.drawable,gGc,x,y,renderX,y);
		
      ++y;
      
      XSetForeground(pDisp,gGc,PALETTE_WHITE);
      XDrawLine(pDisp,lpmo->data.draw.drawable,gGc,x,y,renderX,y);
    }
    else			/* normal item, draw it */
    {
      /* calculate the imageY now */
      if (pParam->icon.icon != None)
      {
	imageY = (lpmo->data.draw.height - pParam->icon.height)/2;
	/* imageBorder = 1; */
      }
      else
      {
	imageBorder = 0;
	imageY = 0;
      }
      
      len = strlen(pParam->name);
      
      /* setup the colors */
      XSetForeground(pDisp,gGc,
		     (lpmo->data.draw.drawFlags & MENU_DF_SELECTED) ? 
		     PALETTE_MENU_BLUE : 
		     PALETTE_LIGHT_GRAY);
      
      pFont->SetColor(pFont,(lpmo->data.draw.drawFlags & MENU_DF_SELECTED) ?
		      PALETTE_WHITE :
		      PALETTE_BLACK);
      /* draw background */
      if ((pParam->icon.icon != None) && 
	  (pParam->ms == MS_FANCY) &&
	  (lpmo->data.draw.drawFlags & MENU_DF_SELECTED))
      {
	renderX = pParam->icon.width + 2*(1+imageBorder) + 2;
	x = lpmo->data.draw.x + renderX;
	y = lpmo->data.draw.width - renderX;
      }
      else
      {
	x = lpmo->data.draw.x;
	y = lpmo->data.draw.width;
      }
      
      XFillRectangle(pDisp,lpmo->data.draw.drawable,gGc,
		     x,lpmo->data.draw.y,
		     y,lpmo->data.draw.height);
      /* now draw icon */
      if (pParam->icon.icon != None)
      {
	/* now render fancy border */
	if ((pParam->ms == MS_FANCY) && 
	    (lpmo->data.draw.drawFlags & MENU_DF_SELECTED))
	{
	  XSetForeground(pDisp,gGc,PALETTE_WHITE);
	  XDrawLine(pDisp,lpmo->data.draw.drawable,gGc,
		    lpmo->data.draw.x,
		    lpmo->data.draw.y + imageY - imageBorder - 1,
		    lpmo->data.draw.x + 2*imageBorder + pParam->icon.width,
		    lpmo->data.draw.y + imageY - imageBorder - 1);
	  XDrawLine(pDisp,lpmo->data.draw.drawable,gGc,
		    lpmo->data.draw.x,
		    lpmo->data.draw.y + imageY - imageBorder - 1,
		    lpmo->data.draw.x,
		    lpmo->data.draw.y + imageY + imageBorder + pParam->icon.height);
	  
	  XSetForeground(pDisp,gGc,PALETTE_DARK_GRAY);
	  XDrawLine(pDisp,lpmo->data.draw.drawable,gGc,
		    lpmo->data.draw.x + 2*imageBorder + pParam->icon.width + 1,
		    lpmo->data.draw.y + imageY - imageBorder - 1,
		    lpmo->data.draw.x + 2*imageBorder + pParam->icon.width + 1,
		    lpmo->data.draw.y + imageY + imageBorder + pParam->icon.height);
	  
	  XDrawLine(pDisp,lpmo->data.draw.drawable,gGc,
		    lpmo->data.draw.x,
		    lpmo->data.draw.y + imageY + imageBorder + pParam->icon.height,
		    lpmo->data.draw.x + 2*imageBorder + pParam->icon.width + 1,
		    lpmo->data.draw.y + imageY + imageBorder + pParam->icon.height); 
	}
	/* render the pixmap */
	RenderPixmap(pDisp,lpmo->data.draw.drawable,gGc,
		     lpmo->data.draw.x + 1 + imageBorder,
		     lpmo->data.draw.y + imageY,
		     &pParam->icon);
	
	/* now calculate where we render the text */
	renderX = pParam->icon.width + 2*(1+imageBorder) + MENU_OFFSET;
      }
      else
      {
	renderX = MENU_OFFSET;
      }
      /* now draw text */
      pFont->GetCenter(pFont,pParam->name,len,
		       lpmo->data.draw.width,lpmo->data.draw.height,&x,&y);
      
      pFont->Render(pFont,lpmo->data.draw.drawable,
		    lpmo->data.draw.x+renderX,lpmo->data.draw.y+y,
		    pParam->name,len);
      /* if submenu, draw arrow */
      if (lpmo->data.draw.drawFlags & MENU_DF_SUBMENU)
      {
	XPoint pts[3];
	
	pts[0].x = lpmo->data.draw.width - TRIANGLE_WIDTH - 2;
	pts[0].y = lpmo->data.draw.y + (lpmo->data.draw.height - TRIANGLE_HEIGHT)/2;
	pts[1].x = lpmo->data.draw.width - TRIANGLE_WIDTH - 2;
	pts[1].y = lpmo->data.draw.y + (lpmo->data.draw.height + TRIANGLE_HEIGHT)/2;
	pts[2].x = lpmo->data.draw.width - 2;
	pts[2].y = lpmo->data.draw.y + lpmo->data.draw.height/2;
	XSetForeground(pDisp,gGc,lpmo->data.draw.drawFlags & MENU_DF_SELECTED ?
		       PALETTE_WHITE : PALETTE_BLACK);
	XFillPolygon(pDisp,lpmo->data.draw.drawable,gGc,pts,3,Nonconvex,CoordModeOrigin);
      }
      /* we're done */
    }
    break;
  case MENU_OP_DESTROY:		
    if (pParam->freeIcon)
    {
      FreePixmap(&pParam->icon);
    }
    if (pParam->name)
    {
      free(pParam->name);
    }
    if (pParam->args)
    {
      free(pParam->args);
    }
    free(pParam);
    break;
  }
}

static void LoadMenuImage(LPSMMI psmi,const char *pFile)
{
  if (strcmp(pFile,"default") == 0)
  {
    psmi->freeIcon = 0;
    memcpy(&psmi->icon,&defaultPixmap,sizeof(PIXMAP));
  }
  else
  {
    psmi->freeIcon = 1;
    LoadPixmapFromFile(pFile,&psmi->icon);
  }
}

static void AddSeparator(LPMENU pProgramMenu)
{
  LPSMMI psmi = (LPSMMI)malloc(sizeof(SMMI));
  if (psmi)
  {
    memset(psmi,0,sizeof(SMMI));
    AddMenuItem(pProgramMenu,(MENUCALLBACK)LaunchProgram,psmi,MENUITEM_SEPARATOR);
  }
}

static void AddProgramMenuItem(LPMENU pProgramMenu,const char *name,const char *args,const char *pIcon,LPPROGRAMMENUOPTIONS lppmop)
{
  LPSMMI psmi = (LPSMMI)malloc(sizeof(SMMI));
  if (psmi)
  {
    memset(psmi,0,sizeof(SMMI));
    psmi->name = (char*)strdup(name);
    psmi->args = (char*)strdup(args);
    psmi->useIcons = lppmop->useIcons;
    psmi->ms = lppmop->ms;
    if (psmi->useIcons && pIcon)
    {
      LoadMenuImage(psmi,pIcon);
    }
    AddMenuItem(pProgramMenu,(MENUCALLBACK)LaunchProgram,psmi,MENUITEM_NORMAL);
  }
}

static LPMENU CreateProgramSubmenu(LPMENU pProgramMenu,const char *name,const char *pIcon,LPPROGRAMMENUOPTIONS lppmop)
{
  LPSMMI psmi = (LPSMMI)malloc(sizeof(SMMI));
  if (psmi)
  {
    memset(psmi,0,sizeof(SMMI));
    psmi->name = (char*)strdup(name);
    psmi->args = NULL;
    psmi->useIcons = lppmop->useIcons;
    psmi->ms = lppmop->ms;
    if (psmi->useIcons && pIcon)
    {
      LoadMenuImage(psmi,pIcon);
    }
    return AddSubMenu(pProgramMenu,(MENUCALLBACK)LaunchProgram,psmi);
  }
  return NULL;
}

static LPMENU ConfigListCreateProgramMenu(LPCONFIGLIST pList,LPMENU pCurrentMenu,LPPROGRAMMENUOPTIONS lppmop)
{
  MENUOPTIONS mop;
  LPCONFIGITEM pConfig;
  LPCONFIGITEM pIcon;
  LPCONFIGITEM pCmd;

  if (!pList)
  {
    return pCurrentMenu;
  }

  if (pCurrentMenu == NULL)
  {
    /* create a menu */
    mop.parent = root;
    mop.background = PALETTE_LIGHT_GRAY;
    pCurrentMenu = CreateMenu(&mop);
  }

  if (pCurrentMenu)
  {
    for (pConfig = (LPCONFIGITEM)pList->list.pHead;pConfig;pConfig=(LPCONFIGITEM)pConfig->node.pNext)
    {
      if (pConfig->value && strcmp(pConfig->name,"icon"))
      {
	if (strcmp(pConfig->name,"separator") == 0)
	{
	  AddSeparator(pCurrentMenu);
	}
	else
	{
	  AddProgramMenuItem(pCurrentMenu,pConfig->name,pConfig->value,NULL,lppmop);
	}
      }
      else if (pConfig->pSublist)
      {
	/* check to see if the child has a cmd item */
	pCmd = FindConfigItemByName(pConfig->pSublist,"cmd");
	pIcon = FindConfigItemByName(pConfig->pSublist,"icon");
	if (pCmd)
	{
	  AddProgramMenuItem(pCurrentMenu,pConfig->name,pCmd->value,
			     pIcon ? pIcon->value : NULL,lppmop);
	}
	else 			/* no cmd, must be submenu */
	{
	  ConfigListCreateProgramMenu(pConfig->pSublist,
				      CreateProgramSubmenu(pCurrentMenu,
							   pConfig->name,
							   pIcon ? pIcon->value :  NULL,lppmop),
				      lppmop);
	}
      }
    }
  }
  return pCurrentMenu;
}

LPMENU CreateProgramMenu(LPPROGRAMMENUOPTIONS lppmop)
{
  LPCONFIGITEM pItem = FindConfigItemByName(lppmop->pList,"start_menu");
  if (pItem)
  {
    return ConfigListCreateProgramMenu(pItem->pSublist,NULL,lppmop);
  }
  return NULL;
}
